# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64decode
from io import BytesIO
from struct import unpack

from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType
from frugal.transport.base_http_transport import TBaseHttpTransport


class TUrlfetchTransport(TBaseHttpTransport):
    """
    Synchronous transport implemented with urlfetch.
    """

    def __init__(self, url, headers=None, get_headers=None):
        """Initialize a new FUrlfetchTransport.

        Args:
            url: url of the Frugal server.
            headers: dict containing static headers.
            get_headers: func which returns dynamic headers per request.
        """

        self._timeout = None
        super(TUrlfetchTransport, self).__init__(url, headers=headers,
                                                 get_headers=get_headers)

    def set_timeout(self, timeout):
        """
        Set the request timeout.

        Args:
            timeout: request timeout in milliseconds.
        """

        self._timeout = timeout / 1000

    def flush(self):
        headers, body = self._get_headers_and_body()

        if not body:
            return

        resp = _urlfetch(self._url, body, self._url.startswith('https://'),
                         self._timeout, headers)

        code = resp.status_code
        if code >= 400:
            msg = 'urlfetch request failed, returned {0}'.format(code)
            raise TTransportException(TTransportExceptionType.UNKNOWN, msg)

        resp_body = b64decode(resp.content)
        # All responses should be framed with 4 bytes (uint32).
        if len(resp_body) < 4:
            msg = 'Invalid frame size.'
            raise TTransportException(TTransportExceptionType.UNKNOWN, msg)

        # If there are only 4 bytes, this needs to be a one-way (i.e. frame
        # size 0)
        if len(resp_body) == 4:
            if unpack('!I', resp_body)[0] != 0:
                msg = 'invalid frame'
                raise TTransportException(TTransportExceptionType.UNKNOWN, msg)

            # It's a oneway, drop it.
            return

        self._rbuff = BytesIO(resp_body[4:])


def _urlfetch(url, body, validate_certificate, timeout, headers):
    from google.appengine.api import urlfetch
    from google.appengine.api.urlfetch_errors import DeadlineExceededError
    try:
        return urlfetch.fetch(
            url, method=urlfetch.POST, payload=body, headers=headers,
            validate_certificate=url.startswith('https://'),
            deadline=timeout
        )
    except DeadlineExceededError:
        raise TTransportException(type=TTransportExceptionType.TIMED_OUT)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from thrift.transport.TTransport import TTransportException, TMemoryBuffer
from tornado import gen

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.transport import FPublisherTransportFactory
from frugal.transport import FPublisherTransport
from frugal.transport import FSubscriberTransportFactory
from frugal.transport import FSubscriberTransport
from frugal.exceptions import TTransportExceptionType

_FRAME_BUFFER_SIZE = 5
_FRUGAL_PREFIX = "frugal."

logger = logging.getLogger(__name__)


class FNatsPublisherTransportFactory(FPublisherTransportFactory):
    def __init__(self, nats_client):
        self._nats_client = nats_client

    def get_transport(self):
        return FNatsPublisherTransport(self._nats_client)


class FNatsPublisherTransport(FPublisherTransport):
    def __init__(self, nats_client):
        super(FNatsPublisherTransport, self).__init__(_NATS_MAX_MESSAGE_SIZE)
        self._nats_client = nats_client

    @gen.coroutine
    def open(self):
        if not self._nats_client.is_connected:
            raise TTransportException(
                type=TTransportExceptionType.NOT_OPEN,
                message="Nats not connected!")

    @gen.coroutine
    def close(self):
        if not self.is_open():
            return

        yield self._nats_client.flush()

    def is_open(self):
        return self._nats_client.is_connected

    @gen.coroutine
    def publish(self, topic, data):
        if not self.is_open():
            raise TTransportException(
                type=TTransportExceptionType.NOT_OPEN,
                message='Nats not connected!')
        if self._check_publish_size(data):
            msg = 'Message exceeds NATS max message size'
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message=msg)
        yield self._nats_client.publish('frugal.{0}'.format(topic), data)


class FNatsSubscriberTransportFactory(FSubscriberTransportFactory):
    def __init__(self, nats_client, queue=''):
        self._nats_client = nats_client
        self._queue = queue

    def get_transport(self):
        return FNatsSubscriberTransport(self._nats_client, self._queue)


class FNatsSubscriberTransport(FSubscriberTransport):
    def __init__(self, nats_client, queue):
        self._nats_client = nats_client
        self._queue = queue
        self._is_subscribed = False
        self._sub_id = None

    @gen.coroutine
    def subscribe(self, topic, callback):
        if not self._nats_client.is_connected:
            raise TTransportException(
                type=TTransportExceptionType.NOT_OPEN,
                message="Nats not connected!")

        if self.is_subscribed():
            raise TTransportException(
                type=TTransportExceptionType.ALREADY_OPEN,
                message="Already subscribed to nats topic!")

        self._sub_id = yield self._nats_client.subscribe_async(
            'frugal.{0}'.format(topic),
            queue=self._queue,
            cb=lambda message: callback(TMemoryBuffer(message.data[4:]))
        )
        self._is_subscribed = True

    @gen.coroutine
    def unsubscribe(self):
        if not self.is_subscribed():
            return

        yield self._nats_client.unsubscribe(self._sub_id)
        self._sub_id = None
        self._is_subscribed = False

    def is_subscribed(self):
        return self._is_subscribed and self._nats_client.is_connected



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct

from thrift.Thrift import TApplicationException
from thrift.transport.TTransport import TMemoryBuffer
from tornado import gen

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.server import FServer
from frugal.transport import TMemoryOutputBuffer

logger = logging.getLogger(__name__)


class FNatsServer(FServer):
    """
    An implementation of FServer which uses NATS as the underlying transport.
    Clients must connect with the FNatsTransport.
    """

    def __init__(self, nats_client, subjects, processor,
                 protocol_factory, queue=""):
        """
        Create a new instance of FStatelessNatsTornadoServer

        Args:
            nats_client: connected instance of nats.io.Client
            subject: subject to listen on
            processor: FProcess
            protocol_factory: FProtocolFactory
        """
        self._nats_client = nats_client
        self._subjects = [subjects] if isinstance(subjects, basestring) \
            else subjects
        self._processor = processor
        self._iprot_factory = protocol_factory
        self._oprot_factory = protocol_factory
        self._queue = queue
        self._sub_ids = []

    @gen.coroutine
    def serve(self):
        """
        Subscribe to provided subject and listen on provided queue
        """
        queue = self._queue
        _cb = self._on_message_callback

        self._sub_ids = [
            (yield self._nats_client.subscribe_async(
                subject,
                queue=queue,
                cb=_cb
            )) for subject in self._subjects
        ]

        logger.info("Frugal server running...")

    @gen.coroutine
    def stop(self):
        """
        Unsubscribe from server subject
        """
        logger.debug("Frugal server stopping...")
        for sid in self._sub_ids:
            yield self._nats_client.unsubscribe(sid)

    @gen.coroutine
    def _on_message_callback(self, msg):
        """
        Process and respond to server request on server subject

        Args:
            msg: request message published to server subject
        """
        reply_to = msg.reply
        if not reply_to:
            logger.warn("Discarding invalid NATS request (no reply)")
            return

        frame_size = struct.unpack('!I', msg.data[:4])[0]
        if frame_size > _NATS_MAX_MESSAGE_SIZE - 4:
            logger.warning("Invalid frame size, dropping message.")
            return

        # Read and process frame (exclude first 4 bytes which
        # represent frame size).
        iprot = self._iprot_factory.get_protocol(
            TMemoryBuffer(msg.data[4:])
        )
        otrans = TMemoryOutputBuffer(_NATS_MAX_MESSAGE_SIZE)
        oprot = self._oprot_factory.get_protocol(otrans)

        try:
            yield self._processor.process(iprot, oprot)
        except TApplicationException:
            # Continue so the exception is sent to the client
            pass
        except Exception:
            return

        if len(otrans) == 4:
            return

        yield self._nats_client.publish(reply_to, otrans.getvalue())



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thrift.transport.TTransport import TTransportBase
from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType


class FTransport(object):
    """
    FTransport is comparable to Thrift's TTransport in that it represent the
    transport layer for frugal clients. However, frugal is callback based and
    sends only framed data. Therefore, instead of exposing read, write, and
    flush, the transport has a simple request method that sends framed frugal
    messages and returns the response.
    """
    def __init__(self, request_size_limit=0):
        """
        Initialize the transport optionally passing a request size limit.

        Args:
            request_size_limit: The maximum request payload size for this
                                transport. A non-positive number indicates
                                an unbounded allowable size.
        """
        self._request_size_limit = request_size_limit

    def open(self):
        """
        Open the transport.
        """
        raise NotImplementedError('You must override this')

    def close(self):
        """
        Close the transport.
        """
        raise NotImplementedError('You must override this')

    def is_open(self):
        """
        Return True if the transport is open, False otherwise.
        """
        raise NotImplementedError('You must override this')

    def set_monitor(self, monitor):
        """
        Set the transport monitor for the transport. This should only be used
        for "stateful" transports.

        Args:
            monitor: A transport monitor
        """
        raise NotImplementedError('You must override this')

    def oneway(self, context, payload):
        """
        Send the given framed frugal payload over the transport.

        Args:
            context:    FContext associated with the request (used for timeout
                        and logging)

            payload:    framed frugal data
        """
        raise NotImplementedError('You must override this')

    def request(self, context, payload):
        """
        Send the given framed frugal payload over the transport and returns the
        response.

        Args:
            context:    FContext associated with the request (used for timeout
                        and logging)

            payload:    framed frugal data
        """
        raise NotImplementedError('You must override this')

    def get_request_size_limit(self):
        """
        Returns the maximum request payload size for this transport. A non-
        positive number is returned to indicate an unbounded allowable size.
        """
        return self._request_size_limit

    def _preflight_request_check(self, payload):
        """
        Helper function that throws TTransportExceptionType.NOT_OPEN if the
        transport is not open or throws FMessageSizeException if the payload is
        too large. Should only be called by extending classes.
        """
        if not self.is_open():
            raise TTransportException(
                type=TTransportExceptionType.NOT_OPEN,
                message='Transport is not open')

        if len(payload) > self.get_request_size_limit() > 0:
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message='Message exceeds {0} bytes, was {1} bytes'.format(
                    self.get_request_size_limit(), len(payload)))


class TSynchronousTransport(TTransportBase, object):
    """
    TSynchronousTransport is a Thrift TTransport for services which make
    synchronous requests.
    """

    def set_timeout(self, timeout):
        """
        Set the request timeout.

        Args:
            timeout: request timeout in milliseconds.
        """
        pass



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tornado import gen

from frugal.transport import FTransport


class FTransportBase(FTransport):
    """
    FTransportBase extends FTransport using the coroutine decorators used by
    all tornado FTransports.
    """
    def is_open(self):
        raise NotImplementedError("You must override this.")

    @gen.coroutine
    def open(self):
        raise NotImplementedError("You must override this.")

    @gen.coroutine
    def close(self):
        raise NotImplementedError("You must override this.")

    @gen.coroutine
    def oneway(self, context, payload):
        raise NotImplementedError('You must override this.')

    @gen.coroutine
    def request(self, context, payload):
        raise NotImplementedError('You must override this.')



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64decode
from io import BytesIO
from struct import unpack

import requests
from requests.exceptions import ReadTimeout
from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType
from frugal.transport.base_http_transport import TBaseHttpTransport


class THttpTransport(TBaseHttpTransport):
    """
    Synchronous transport implemented with Requests.
    """

    def __init__(self, url, request_capacity=0, response_capacity=0,
                 headers=None, get_headers=None):
        """
        Initialize a new THttpTransport.

        Args:
            url: url of the Frugal server.
            request_capacity: max size allowed to be written in a request. Set
                              0 for no restriction.
            response_capacity: max size allowed to be read in a response. Set
                               0 for no restriction.
            headers: dict containing static headers.
            get_headers: func which returns dynamic headers per request.
        """

        super(THttpTransport, self).__init__(
            url, request_capacity=request_capacity,
            response_capacity=response_capacity, headers=headers,
            get_headers=get_headers)

    def flush(self):
        headers, body = self._get_headers_and_body()

        if not body:
            return

        timeout = None
        if self._timeout:
            # Requests uses timeout in seconds.
            timeout = self._timeout / 1000.0
            if timeout <= 0:
                timeout = None

        try:
            resp = requests.post(self._url, data=body, headers=headers,
                                 timeout=timeout)
        except ReadTimeout:
            raise TTransportException(
                type=TTransportExceptionType.TIMED_OUT,
                message='Request timed out')
        if resp.status_code == 413:
            raise TTransportException(
                type=TTransportExceptionType.RESPONSE_TOO_LARGE,
                message='response was too large for the transport'
            )
        if resp.status_code >= 400:
            raise TTransportException(
                type=TTransportExceptionType.UNKNOWN,
                message='HTTP request failed, returned {0}: {1}'.format(
                    resp.status_code, resp.reason))

        resp_body = b64decode(resp.content)
        # All responses should be framed with 4 bytes (uint32).
        if len(resp_body) < 4:
            raise TTransportException(
                type=TTransportExceptionType.UNKNOWN,
                message='invalid frame size')

        # If there are only 4 bytes, this needs to be a one-way (i.e. frame
        # size 0)
        if len(resp_body) == 4:
            if unpack('!I', resp_body)[0] != 0:
                raise TTransportException(
                    type=TTransportExceptionType.UNKNOWN,
                    message='invalid frame')

            # It's a oneway, drop it.
            return

        self._rbuff = BytesIO(resp_body[4:])



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64encode
from base64 import b64decode
import logging
from struct import pack

from six.moves import BaseHTTPServer
from thrift.Thrift import TApplicationException
from thrift.transport import TTransport

from frugal.server import FServer

logger = logging.getLogger(__name__)


class FHttpServer(FServer):
    """
    Simple FServer implementation using HTTP. This is merely a reference
    implementation and is not production-ready.
    """

    def __init__(self, processor, address, proto_factory):
        """
        Initialize an FHttpServer.

        Args:
            processor: FProcessor used to process requests.
            address: tuple of host name and port.
            proto_factory: FProtocolFactory used to read requests and write
                           responses.
        """

        class RequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
            """
            Request handler implements a frugal HTTP handler.
            """

            def do_POST(self):
                """
                do_POST implements POST method on a frugal server.
                """
                length = self.headers.getheader('Content-Length')
                payload = b64decode(self.rfile.read(int(length)))
                response_limit = int(self.headers.getheader(
                    'x-frugal-payload-limit', 0))

                if len(payload) <= 4:
                    logging.exception(
                        'Invalid request frame length %s', len(payload))
                    self.send_response(400)
                    self.end_headers()
                    return

                itrans = TTransport.TMemoryBuffer(payload[4:])
                otrans = TTransport.TMemoryBuffer()
                iprot = proto_factory.get_protocol(itrans)
                oprot = proto_factory.get_protocol(otrans)

                try:
                    processor.process(iprot, oprot)
                except TApplicationException:
                    # Continue so the exception is sent to the client
                    pass
                except Exception:
                    self.send_response(400)
                    self.end_headers()
                    return

                # Encode response.
                response = otrans.getvalue()
                if len(response) > response_limit > 0:
                    self.send_response(413)
                    self.end_headers()
                    return

                frame_length = pack('!I', len(response))
                frame = b64encode(frame_length + response)

                self.send_response(200)
                self.send_header('Content-Type', 'application/x-frugal')
                self.send_header('Content-Length', len(frame))
                self.send_header('Content-Transfer-Encoding', 'base64')
                self.end_headers()

                self.wfile.write(frame)

        self._httpd = BaseHTTPServer.HTTPServer(address, RequestHandler)

    def serve(self):
        self._httpd.serve_forever()

    def stop(self):
        self._httpd.socket.close()



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from frugal.protocol import FProtocol


class FProtocolFactory(object):
    """FProtocolFactory creates FProtocols."""

    def __init__(self, t_protocol_factory):
        """Initialize FProtocolFactory.

        Args:
            t_protocol_factory: Thrift TProtocolFactory.
        """
        self._t_protocol_factory = t_protocol_factory

    def get_protocol(self, transport):
        """Return FProtocol for the given TTransport.

        Args:
            transport: Thrift TTransport.

        Returns:
            FProtocol wrapping the TTransport.
        """
        return FProtocol(self._t_protocol_factory.getProtocol(transport))



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six


def make_hashable(thing):
    """
    Creates a representation of some input data with immutable collections,
    so the output is hashable.

    :param thing: The data to create a hashable representation of.
    """
    if isinstance(thing, list):
        return tuple(make_hashable(elem) for elem in thing)
    elif isinstance(thing, set):
        return frozenset(make_hashable(elem) for elem in thing)
    elif isinstance(thing, dict):
        return frozenset((make_hashable(k), make_hashable(v))
                         for k, v in six.iteritems(thing))
    else:
        return thing



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Lock

from thrift.transport.TTransport import TTransportException
from thrift.transport.TTransport import TFramedTransport

from . import FServer


class FSimpleServer(FServer):
    """
    FSimpleServer is a very basic implementation of an FServer.
    """

    def __init__(self, processor_factory, server_transport, protocol_factory):
        self._processor_factory = processor_factory
        self._server_transport = server_transport
        self._protocol_factory = protocol_factory
        self._stopped = False
        self._stopped_mu = Lock()

    def serve(self):
        self._server_transport.listen()
        while True:
            with self._stopped_mu:
                if self._stopped:
                    return

            client = self._server_transport.accept()
            framed = TFramedTransport(client)
            iprot = self._protocol_factory.get_protocol(framed)
            oprot = self._protocol_factory.get_protocol(framed)
            processor = self._processor_factory.get_processor(framed)

            try:
                while True:
                    with self._stopped_mu:
                        if self._stopped:
                            break

                    processor.process(iprot, oprot)
            except TTransportException:
                continue
            except Exception as e:
                print(e)
                break

            framed.close()

    def stop(self):
        self._stopped = True



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging

from thrift.Thrift import TApplicationException
from thrift.Thrift import TMessageType
from thrift.Thrift import TType
from tornado import gen
from tornado.locks import Lock

from frugal.exceptions import TApplicationExceptionType

logger = logging.getLogger(__name__)


class FProcessorFunction(object):
    """
    FProcessorFunction is a generic object that exposes a single process
    call, which is used to handle a method invocation. FProcessorFunction
    should be implemented by the generated code.
    """

    def __init__(self, handler, lock):
        """
            Args:
                handler: frugal.middleware.Method
                lock: tornado.locks.Lock
        """
        self._handler = handler
        self._lock = lock

    @gen.coroutine
    def process(self, ctx, iprot, oprot):
        """
        Process the request from the input protocol and write the
        response to the output protocol.

        Args:
            iprot: input FProtocol
            oprot: output FProtocol
        """
        pass

    def add_middleware(self, middleware):
        """
        Add the given middleware to the FProcessorFunction
        This should only be called before the server is started.

            Args:
             middleware: ServiceMiddleware
         """

        self._handler.add_middleware(middleware)


class FProcessor(object):
    """
    FProcessor is a generic object which operates upon an input stream and
    writes to some output stream.
    """

    @gen.coroutine
    def process(self, iprot, oprot):
        """
        Process the request from the input protocol and write the
        response to the output protocol.

        Args:
            iprot: input FProtocol
            oprot: output FProtocol
        """
        pass

    def add_middleware(self, middleware):
        """
        Add the given middleware to the FProcessor
        This should only be called before the server is started.

            Args:
             middleware: ServiceMiddleware
         """
        pass

    def get_annotations_map(self):
        """
        Return a deepcopy of the annotations map.
        """
        pass


class FBaseProcessor(FProcessor):
    """
    FBaseProcessor is a base implementation of FProcessor. FProcessors
    should extend this and map FProcessorFunctions. This should only be used
    by generated code.
    """

    def __init__(self):
        """
        Create new instance of FBaseProcessor that will process requests.
        """
        self._processor_function_map = {}
        self._annotations_map = {}
        self._write_lock = Lock()

    def add_to_processor_map(self, key, proc):
        """
        Register the given FProcessorFunction.

        Args:
            key: processor function name
            proc: FProcessorFunction
        """
        self._processor_function_map[key] = proc

    def add_to_annotations_map(self, method_name, annotation):
        """
        Register the given annotation dictionary

        Args:
            method_name: method name
            annotation: annotation dictionary
        """
        self._annotations_map[method_name] = annotation

    def get_annotations_map(self):
        """
        Return a deepcopy of the annotations map.
        """
        return copy.deepcopy(self._annotations_map)

    def get_write_lock(self):
        """
        Return the write lock.
        """
        return self._write_lock

    @gen.coroutine
    def process(self, iprot, oprot):
        """
        Process an input protocol and output protocol

        Args:
            iprot: input FProtocol
            oport: ouput FProtocol

        Raises:
            TException: if processing fails.
        """
        context = iprot.read_request_headers()
        name, _, _ = iprot.readMessageBegin()

        processor_function = self._processor_function_map.get(name)

        # If the function was in our dict, call process on it.
        if processor_function:
            try:
                yield processor_function.process(context, iprot, oprot)
            except Exception:
                # Don't raise an exception because the server should still send
                # a response to the client.
                logger.exception(
                    "frugal: exception occurred while processing request with "
                    "correlation id %s", context.correlation_id)
            return

        logger.warn("frugal: client invoked unknown method %s on request "
                    "with correlation id %s", name, context.correlation_id)
        iprot.skip(TType.STRUCT)
        iprot.readMessageEnd()

        ex = TApplicationException(
            type=TApplicationExceptionType.UNKNOWN_METHOD,
            message="Unknown function: {0}".format(name))

        with (yield self._write_lock.acquire()):
            oprot.write_response_headers(context)
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, 0)
            ex.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()

    def add_middleware(self, middleware):
        """
        Add the given middleware to the FProcessor.
        This should only be called before the server is started.

        Args:
            middleware: ServiceMiddleware
        """

        if middleware and not isinstance(middleware, list):
            middleware = [middleware]

        for proc in self._processor_function_map.values():
            proc.add_middleware(middleware)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FSubscription(object):
    """
    FSubscription to a pub/sub topic.  This is used only by generated code
    and should not be called directly.
    """

    def __init__(self, topic, transport):
        """
        Initialize FSubscription.

        Args:
            topic: pub/sub topic string.
            transport: FScopeTransport for the subscription.
        """
        self._topic = topic
        self._transport = transport

    def get_topic(self):
        """
        Return subscription topic.
        """
        return self._topic

    def unsubscribe(self):
        """
        Unsubscribe from the topic.

        The result of this is a future that should be awaited/yielded
        appropriately.
        """
        return self._transport.unsubscribe()

    def remove(self):
        """
        Unsubscribe and removes durably stored information on the broker,
        if applicable.
        """
        return self._transport.remove()



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nats.aio.client import Client
from nats.aio.utils import new_inbox
from thrift.transport.TTransport import TTransportException

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.aio.transport import FAsyncTransport
from frugal.exceptions import TTransportExceptionType


class FNatsTransport(FAsyncTransport):
    """
    FNatsTransport is an extension of FAsyncTransport that uses nats as the
    underlying transport. This is "stateless" in the sense there is no
    connection with a server. A request is published on a subject and responses
    are received on another subject. To use this, requests and responses MUST
    fit within a single nats message.
    """
    def __init__(
        self,
        nats_client: Client,
        subject: str,
        inbox=''
    ):
        super().__init__(request_size_limit=_NATS_MAX_MESSAGE_SIZE)
        self._nats_client = nats_client
        self._subject = subject
        self._inbox = inbox or new_inbox()
        self._is_open = False
        self._sub_id = None

    def is_open(self):
        """Check whether the transport is open."""
        return self._is_open and self._nats_client.is_connected

    async def open(self):
        """Subscribe to the inbox subject."""
        if not self._nats_client.is_connected:
            raise TTransportException(TTransportExceptionType.NOT_OPEN,
                                      'Nats not connected')

        if self.is_open():
            raise TTransportException(TTransportExceptionType.ALREADY_OPEN,
                                      'Transport is already open')

        self._sub_id = await self._nats_client.subscribe_async(
            self._inbox,
            cb=self._on_message_callback
        )
        self._is_open = True

    async def _on_message_callback(self, message):
        await self.handle_response(message.data[4:])

    async def close(self):
        """Unsubscribe from the inbox subject."""
        if not self._sub_id:
            return

        await self._nats_client.unsubscribe(self._sub_id)
        self._is_open = False
        self._sub_id = None

    async def flush(self, data):
        await self._nats_client.publish_request(
            self._subject,
            self._inbox,
            data
        )
